// (c) Copyright 2011-2012 Cloudera, Inc. All rights reserved.
define([
  'cloudera/EventSearch',
  'cloudera/common/Humanize',
  'cloudera/Util',
  'cloudera/common/TimeUtil',
  'cloudera/chart/TimeBrowserState',
  'underscore',
  // Below this line, we only depend on the modules to be loaded,
  // not use them by name.
  'cloudera/MegaBase',
  'cloudera/knockout/ko.formattedDate',
  'cloudera/knockout/ko.joinText',
  'cloudera/knockout/ko.select2'
], function(EventSearch, Humanize, Util, TimeUtil, TimeBrowserState, _) {

    var HistorySearch = function(filterElement, listElement, historyUrl, filtersUrl, useTimeControl) {
        var self = this;
            // The 'state' variable is used to identify a state of the filter
            // list to ensure that we don't load data that is no longer applicable.
        var state = 0,
            // Static filters can be embedded in a template to limit the filterable result. This is used on the service audit pages.
            staticFilters = $(filterElement).data('static-filters');
        // Base parameters to be passed to the 'history' service.
        self.scmUrlParams = {
            filters: staticFilters,
            offset: 0,
            limit: 50
        };

        /**
         * Initiates a download of a CSV file.
         */
        self.exportEvents = function() {
          var params = [],
            param,
            exportParams = $.extend({}, self.scmUrlParams, {
                limit: -1,
                format: "CSV",
                attachment: true
            });
          if (exportParams.filters) {
            exportParams.filters = JSON.stringify(exportParams.filters);
          }
          _.each(exportParams, function(value, key) {
              params.push(encodeURIComponent(key) + "=" + encodeURIComponent(value));
          });
          Util.setWindowLocation(historyUrl + '?' + params.join("&"));
        };

        self.subscriptionHandles = [];

        /**
         * Applies the current selected filters to the data set.
         */
        self.applyFilters = function() {
            self.eventSearch.deactivateAllFilters();
            self.scmUrlParams.filters = self.eventSearch.getAppliedFilters();
            // Incrementing the state to ensure we ignore result from previous filter states.
            state += 1;
            // Clear results and load new the data.
            self.eventSearch.loadData(true);
        };

        // Return a function suitable for use in the later call to retrieve
        // history events from the server.
        self.makeOnRetrieveHistorySuccess = function(clear, eventSearch) {
            // Persist the state at the time of the request.
            var requestState = state,
            // Persist the length of the events at the time on the request.
            initialLength = clear ? 0 : eventSearch.length();
            // Set the requested offset to the current length of the results.
            self.scmUrlParams.offset = initialLength;

            return function(data) {
                var currentLength = eventSearch.length(),
                    searcher,
                    newest,
                    oldest,
                    prependRange;
                // Ignore the result if the current state doesn't equal
                // the state when the request was made
                if (requestState !== state
                        // or there's no data returned
                        || data === null
                        || (!data.length)
                        // or another request beat this request in.
                        || initialLength !== currentLength) {
                    return;
                }

                // If the list is currently empty, just push the results
                // into the list.
                if (!currentLength) {
                    eventSearch.append(data);
                    return;
                }

                oldest = eventSearch.getTail().timestamp;
                // If the oldest record in the list is newer than the newest result append the result
                if (data[0].timestamp < oldest) {
                    eventSearch.append(data);
                    // otherwise, attempt to prepend the result.
                    // (this code should probably be moved into the prepend function of EventSearch.js
                } else {
                    data.reverse();
                    searcher = function(i) {
                        return data[i].timestamp;
                    };

                    newest = eventSearch.getHead().timestamp;
                    prependRange = Util.findRange(searcher, data.length-1, newest+1, self.scmUrlParams.endTime);
                    if (prependRange) {
                        eventSearch.prepend(data.slice(prependRange.start, prependRange.end).reverse());
                    }
                }
            };
        };

        /**
         * This is the primary data handling function for HistorySearch.
         */
        self.dataSource = function(eventSearch, clear, callback) {
          // A variable to store an error during the request phase if applicable.
          // WARNING: makeOnRetrieveHistorySuccess needs to be called before the params are cloned.
          var successHandler = self.makeOnRetrieveHistorySuccess(clear, eventSearch),
              params = _.clone(self.scmUrlParams),
              err;

          // Clear the current results of they're no longer applicable.
          if (clear) {
              eventSearch.clear();
          }

          /* Because of the complexity of the filter model, it is causing $.params to
           * JSON.stringify and then URL encode the object. So this will stringify just
           * the filters object and convert the rest of the object into a query string.
           */
          if (params.filters) {
            params.filters = JSON.stringify(params.filters);
          }

          $.getJSON(historyUrl, params)
              .success(successHandler)
              .error(function(jqXHR, textStatus, errorThrown) {
                  // Store the error for later use.
                  err = errorThrown;
              })
              .complete(function() {
                  // Invoke the callback, passing an error if one occurred.
                  callback(err);
              });
        };

        self.onTimeSelectionChanged = function(range, isCurrentMode, isAutoUpdate) {
            // Ignore auto update.
            if (isAutoUpdate) {
                return;
            }
            self.scmUrlParams.startTime = range.startDate.getTime();
            self.scmUrlParams.endTime = range.endDate.getTime();
            lastIncrement = 0;
            self.eventSearch.loadData(true);
        };

        self.subscribe = function() {
          Util.unsubscribe(self);
          /** Time Control **/
          // Listen for timeSelectionChanged event and request and up-to-date result.
          if (useTimeControl) {
            self.subscriptionHandles = [$.subscribe("timeSelectionChanged", self.onTimeSelectionChanged)];
          } else {
            self.onTimeSelectionChanged({startDate: new Date(0), endDate: TimeUtil.getServerNow()});
          }

          /**
           * Subscribes to the changed event of appliedfilters.
           */
            self.subscriptionHandles.push($.subscribe('eventsearch.appliedfilters.changed', function() {
                self.applyFilters();
            }));
        };

        self.eventSearch = new EventSearch(
            filterElement,
            listElement,
            filtersUrl,
            staticFilters,
            _.bind(self.dataSource, self),
            _.bind(self.exportEvents, self)
        );

        // This is to avoid showing the "No records found..." message while waiting for the context to load.
        self.eventSearch.loading(true);

        // Load the applicable filters from the server.
        self.eventSearch.loadFilters();

        /** Infinite Scroll **/
        var lastIncrement = 0,
            loadingNext = false,
            $history = $('#eventListContainer').scroll(function(e) {
                var position = $history.scrollTop(),
                historyScrollHeight = $history.prop('scrollHeight');
                if (!loadingNext
                        && historyScrollHeight > lastIncrement
                        && (historyScrollHeight - $history.innerHeight() - position) < 200) {
                    loadingNext = true;
                    self.eventSearch.loadData(function(err){
                        if(err) {
                            return true;
                        } else {
                            loadingNext = false;
                        }
                    });
                    lastIncrement = historyScrollHeight;
                }
            });

        /** Filter Activation **/
        $('.event-list').on('click', 'a', function(e){
            var $target = $(e.target), filterName = $target.data('filter'), filter;
            if (filterName) {
                filter = self.eventSearch.findFilter(filterName);
                if (filter) {
                    self.eventSearch.applyFilter(filter, 'EQ', $target.text());
                    self.applyFilters();
                }
            }
        });

        $('#applyFilters').click(self.applyFilters);

        // Handle 'Enter' to apply filters.
        var filterForm = $(filterElement).keypress(function(e) {
            // 13 is the Enter key.
            if (e.which === 13 && $(e.target).is('.text-value')) {
                filterForm.find('input').blur();
                _.defer(_.bind(self.applyFilters, self));
            }
        });

        self.checkInitialTimeRange = function() {
            // If we can find the cached values of the current TC time then
            // we missed the initial timeSelectionChanged event and need to
            // update manually.
            var cachedValues = TimeBrowserState.getTimeSelectionValues();
            if (cachedValues) {
                self.onTimeSelectionChanged(cachedValues.timeRange,
                    cachedValues.isCurrentMode, cachedValues.isAutoUpdate);
            }
        };

        self.subscribe();

        // Check to see if we've missed the initial time range selection event.
        self.checkInitialTimeRange();
    };

    return HistorySearch;
});
